<?php
// $Id: Bootstrap.php,v 1.2 2009/02/24 08:09:40 chop Exp $
/**
* DrupalTest bootstrapping class
*
* This class provides all the functionality for bootstrapping Drupal 
* and also ensuring the test environment is set up before each test 
* and torn down after.
*
* @package DrupalTest
* @author  Jonathan H. Wage <jonwage@gmail.com>
* @author  Christopher Hopper <christopher.jf.hopper@gmail.com>
*/
class DrupalTest_Bootstrap
{
  /**
   * The path to the folder where Drupal is installed. This value is 
   * always set by the class constructor upon successful instantiation.
   * @see DrupalTest_Bootstrap::__construct
   * @var string
   */
  public $drupalRoot;

  /**
   * DrupalTest Bootstrap constructor
   */
  function __construct() {
    /**
     * Find the root (install) path of Drupal or die.
     */
    $testPath = dirname(__FILE__) .'/../../../../';
    for ($pathJumps = 1, $notFound = TRUE; $pathJumps <= 2 && $notFound; $pathJumps++) {
      $this->drupalRoot = realpath($testPath);
      if (file_exists($this->drupalRoot .'/includes/bootstrap.inc')) {
        $notFound = FALSE;
      }
      else {
        $testPath .= '../../';
      }
    }
    if ($notFound) {
      $msg = "bootstrap.inc could not be loaded. Last try at '". $this->drupalRoot ."'". PHP_EOL;
      if (PHP_SAPI == 'cli') {
        fwrite(STDERR, $msg);
        exit(1);
      }
      else {
        exit($msg);
      }
    }
  }

  /**
   * Get instance of bootstrapping class
   *
   * @return DrupalTest_Bootstrap 
   *   Instance of this class.
   */
  public static function getInstance()
  {
    static $instance;

    if (!$instance) {
      $instance = new DrupalTest_Bootstrap();
    }

    return $instance;
  }

  /**
   * Run bootstrapping process
   *
   * @return void
   */
  public function run()
  {  

    $cmd = 'index.php';
    $_SERVER['HTTP_HOST']       = 'default';
    $_SERVER['PHP_SELF']        = '/'. $cmd;
    $_SERVER['REMOTE_ADDR']     = '127.0.0.1';
    $_SERVER['SERVER_SOFTWARE'] = 'PHP '. strtoupper(PHP_SAPI);
    $_SERVER['REQUEST_METHOD']  = 'GET';
    $_SERVER['QUERY_STRING']    = '';
    $_SERVER['PHP_SELF']        = $_SERVER['REQUEST_URI'] = '/';

    $this->drupalBootstrap();

  }

  /**
   * Execute Drupal core bootstrapping function
   *
   * @param mixed $phase 
   * @return void
   */
  public function drupalBootstrap($phase = 0)
  {
    $cwd = getcwd();
    chdir($this->drupalRoot);

    require_once './includes/bootstrap.inc';
    
    if (!$phase) $phase = DRUPAL_BOOTSTRAP_FULL;

    /** 
     * Clean-up $_SERVER['SCRIPT_NAME'] for php-cli SAPI on Windows so 
     * that settings.php can be found correctly.
     * @see conf_path
     */
    if (PHP_SAPI == 'cli' && stripos($_SERVER['OS'], 'windows') !== FALSE) {
      // Switch all back-slashes to forward-slashes in the path.
      $_SERVER['SCRIPT_NAME'] = strtr($_SERVER['SCRIPT_NAME'], "\\", '/');
      // Remove any leading drive letter identifier. 
      if (($scr_idx = strpos($_SERVER['SCRIPT_NAME'], ':')) !== FALSE) {
        $_SERVER['SCRIPT_NAME'] = substr($_SERVER['SCRIPT_NAME'], $scr_idx + 1);
      }
      unset($scr_idx);
    }
    
    drupal_bootstrap($phase);
    
    chdir($cwd);
  }

  /**
   * Backup database 
   * 
   * Used during set-up before each test
   *
   * @todo When a native method for instantiating a Drupal database 
   * exists we should replace this ugly hack with something cleaner and 
   * native.
   *
   * @return void
   */
  public function backupDatabase()
  {
    $info = $this->getTestDbInfo();

    $dumpCommand = 'mysqldump -u ' . $info['user'];
    if ($info['pass'])
    {
      $dumpCommand .= ' -p ' . $info['pass'];
    }

    $dumpCommand .= ' ' . $info['db_name'] . ' > ' . $info['dump_name'];
    exec($dumpCommand);
  }

  /**
   * Restore database 
   * 
   * Used during tear-down to restore from the backup created by 
   * backupDatabase(). This function simply executes a .sql file 
   * generated by mysqldump. 
   *
   * @todo See the todo statement in backupDatabase()
   *
   * @return void
   * @see DrupalTest_Bootstrap::backupDatabase
   */
  public function restoreDatabase()
  {
    $info = $this->getTestDbInfo();

    $restoreCommand = 'mysql -u ' . $info['user'];
    if ($info['pass'])
    {
      $restoreCommand .= ' -p ' . $info['pass'];
    }

    $restoreCommand .= ' ' . $info['db_name'] . ' < ' . $info['dump_name'];
    exec($restoreCommand);      
  }

  /**
   * Get the parsed database information for the drupal environment
   *
   * @return array $info
   */
  public function getTestDbInfo()
  {
    global $db_url;

    if (empty($this->_testDbInfo)) {
      if (is_array($db_url)) {
        $default_db = $db_url['default'];
      }
      else {
        $default_db = $db_url;
      }
      $this->_testDbInfo = parse_url($default_db);
      $this->_testDbInfo['pass'] = isset($this->_testDbInfo['pass']) ? $this->_testDbInfo['pass']:null;
      $this->_testDbInfo['db_name'] = substr($this->_testDbInfo['path'], 1, strlen($this->_testDbInfo['path']));
      $this->_testDbInfo['dump_name'] = '/tmp/' . $this->_testDbInfo['db_name'] . '.original.sql';
    }

    return $this->_testDbInfo;
  }
}